"""Module for imputation performance evaluation

    Function to generate test dataset with missing SNPs, impute missing SNPs
    and compare precision with original dataset.

    In other word, this module provides function to empirically measure imputation R2 on independant dataset.
"""

from raiss.pipes import save_chromosome_imputation
import multiprocessing
from joblib import Parallel, delayed
import pandas as pd
import numpy as np



def generated_test_data(zscore, N_to_mask=5000, condition=None):
    """
        Mask N_to_mask Snps in the dataframe zscore and return the dataframe with missing SNPs.
        Args:
            zscore (pandas DataFrame): Zscores DataFrame
            condition (None or pandas boolean Series): If None, SNPs will be mask randomly. If a pandas boolean series is passed, masked SNPs will be randomly chosen inside the SNPs which have True value.
            N_to_mask (int): Number of SNPs to mask.
    """
    if isinstance(condition, pd.Series)==False:
        masked = np.random.choice(zscore.index, N_to_mask, replace=False)
    else:
        try:
            masked = np.random.choice(zscore.index[condition], N_to_mask, replace=False)
        except ValueError as ve:
            print("Couldn't sample {0} SNPs".format(N_to_mask))
            print("raise {0}".format(ve))

    known = zscore.index.difference(masked)
    return (zscore.loc[known], masked)

def imputation_performance(zscore_initial, zscore_imputed, masked):
    """
    Compute imputation performance on real data
    the performance is computed as the correlation between imputed and real values
    and as the mean absolute deviation between imputed and real values
    Args:
        zscore_initial (pandas DataFrame):
        zscore_imputed (pandas DataFrame):
        masked : SNPs ids which have been masked by imputation
    """
    try:
        fraction_imputed = 1.0-zscore_imputed.loc[masked, "Z"].isnull().mean()
        #cor = zscore_initial.loc[masked, "Z"].corr(zscore_imputed.loc[masked, "Z"].fillna(0))
        cor = zscore_initial.loc[masked, "Z"].corr(zscore_imputed.loc[masked, "Z"])
        MAE = (zscore_initial.loc[masked, "Z"] - zscore_imputed.loc[masked, "Z"]).dropna().abs().mean()
        return {'N_SNP':len(masked),'fraction_imputed':fraction_imputed, 'cor':cor, 'mean_absolute_error':MAE}

    except KeyError: # If KeyError none of the masked_SNP are in the imputed dataframe
        res = np.nan
        return {'N_SNP':np.nan, 'fraction_imputed': np.nan, 'cor':np.nan, 'mean_absolute_error':np.nan}

def z_amplitude_effect(zscore_folder, masked_folder, output_folder, ref_folder, ld_folder, gwas, z_treshold = [0, 1.0, 2.0, 3.0, 4.0, 5], window_size= 500000, buffer_size=125000, eigen_ratio = 0.1, chrom="chr22",  l2_regularization=0.1, R2_threshold=0.6, ratio_to_mask=0.5):
    """
    Compute the imputation performance on SNPs with different amplitude
    The procedure is the following:
    * Masked Snps in the input dataset in function of the amplitude

    for eigen ratio in **eigen_ratio_grid**:
        * Impute SNPs
        * Compute performance on masked SNps
    Args:
        zscore_folder (str): path toward the input data folder
        masked_folder (str) : path toward the folder to save the dataframe with masked SNPs
        output_folder (str) : path toward the folder to save the imputed dataframe
        ref_folder (str) : path toward the folder to save the imputed dataframe
        ld_folder (str) : path toward the Linkage desiquilibrium matrices
         to save the imputed dataframe
        gwas (str): gwas identifier in the following format : 'CONSORTIA_TRAIT'
        chrom (str): chromosome in the format "chr.."
        z_treshold (list) : float list to select Z score to mask above Z score
        eigen_ratio (float): rcond parameter (must be between 0 and 1)
        window_size, buffer_size, l2_regularization, R2_threshold : imputation parameter (see raiss command line documentation)
        ratio_to_mask (int): the fraction of SNPs above the z_threshold to mask in the initial dataset to compute the correlation between true value and imputed value
    """
    z_file = "{0}/z_{1}_{2}.txt".format(zscore_folder, gwas, chrom)
    zscore = pd.read_csv(z_file, index_col=0, sep="\t")
    z_output = "{0}/z_{1}_{2}.txt".format(output_folder, gwas, chrom)
    dat_orig = pd.read_csv(z_file, sep="\t", index_col=0)

    def run_imputation(z, i):
        print("Z score threshold : {}".format(z))
        condition = (dat_orig.Z.abs() > z)
        print("Number of SNPs verifying condition : {}".format(condition.sum()))
        N_to_mask = np.int(np.floor(condition.sum()*ratio_to_mask))

        res_masked = generated_test_data(dat_orig, N_to_mask= N_to_mask, condition = condition)
        z_masked = res_masked[0]
        z_masked_file = "{0}/z_{1}_{2}_{3}.txt".format(masked_folder, gwas, i, chrom)
        z_masked.to_csv(z_masked_file, sep="\t")

        masked_SNP = res_masked[1]
        tag = "_{}".format(z)
        gwas_proc = "{0}_{1}".format(gwas, i)
        save_chromosome_imputation(gwas_proc, chrom, window_size, buffer_size, l2_regularization, eigen_ratio, masked_folder, ref_folder, ld_folder, output_folder, R2_threshold, tag)

        z_output = "{0}/z_{1}_{2}{3}.txt".format(output_folder, gwas_proc, chrom, tag)
        dat_imp = pd.read_csv(z_output, sep="\t", index_col=0)
        R2 = imputation_performance(dat_orig, dat_imp, masked_SNP)
        print("R2 for {0} Zscore z_treshold: {1}".format(z, R2))
        return(R2)

    n_cpu = multiprocessing.cpu_count()
    R2_val = Parallel(n_jobs=n_cpu)(delayed(run_imputation)(z, i) for i,z in enumerate(z_treshold))
    R2_serie =pd.DataFrame(R2_val, columns = ['N_SNP','fraction_imputed','cor', 'mean_absolute_error'], index = z_treshold)
    return(R2_serie)


def grid_search(zscore_folder, masked_folder, output_folder, ref_folder, ld_folder, gwas, chrom="chr22", eigen_ratio_grid = [0.5, 0.1, 0.01], window_size= 500000, buffer_size=125000, l2_regularization=0.1, R2_threshold=0.6, N_to_mask=5000):
    """
    Compute the imputation performance for several eigen ratioself.
    The procedure is the following:
    * Masked N_to_mask Snps in the input dataset (chosen at random)
    * for eigen ratio in **eigen_ratio_grid**:
        * Impute SNPs
        * Compute performance on masked SNPs (R-square and mean absolute error)
    Args:
        zscore_folder (str): path toward the input data folder
        masked_folder (str) : path toward the folder to save the dataframe with masked SNPs
        output_folder (str) : path toward the folder to save the imputed dataframe
        ref_folder (str) : path toward the folder to save the imputed dataframe
        ld_folder (str) : path toward the Linkage desiquilibrium matrices
         to save the imputed dataframe
        gwas (str): gwas identifier in the following format : 'CONSORTIA_TRAIT'
        chrom (str): chromosome in the format "chr.."
        eigen_ratio_grid (list): list of eigen_ratio to test (must be between 0 and 1)
        window_size, buffer_size, l2_regularization, R2_threshold : imputation parameter (see raiss command line documentation)
        N_to_mask (int): Number of SNPs masked in the initial dataset to compute the correlation between true value and imputed value
    """
    z_file = "{0}/z_{1}_{2}.txt".format(zscore_folder, gwas, chrom)

    z_output = "{0}/z_{1}_{2}.txt".format(output_folder, gwas, chrom)
    dat_orig = pd.read_csv(z_file, sep="\t", index_col=0)

    res_masked = generated_test_data(dat_orig, N_to_mask)
    z_masked = res_masked[0]
    z_masked_file = "{0}/z_{1}_{2}.txt".format(masked_folder, gwas, chrom)
    z_masked.to_csv(z_masked_file, sep="\t")
    masked_SNP = res_masked[1]

    def run_imputation(cond):
        tag = "_{}".format(cond)
        save_chromosome_imputation(gwas, chrom, window_size, buffer_size, l2_regularization, cond, masked_folder, ref_folder, ld_folder, output_folder, R2_threshold, tag)

    n_cpu = multiprocessing.cpu_count()
    Parallel(n_jobs=n_cpu)(delayed(run_imputation)(rd) for rd in eigen_ratio_grid)

    R2_serie = pd.DataFrame({"cor":np.nan, "mean_absolute_error":np.nan, "fraction_imputed":np.nan}, index = eigen_ratio_grid)

    for rd in eigen_ratio_grid:
        z_output = "{0}/z_{1}_{2}_{3}.txt".format(output_folder, gwas, chrom, rd)
        dat_imp = pd.read_csv(z_output, sep="\t", index_col=0)
        try:
            res = imputation_performance(dat_orig, dat_imp, masked_SNP)
        except KeyError: # If KeyError none of the masked_SNP are in the imputed dataframe
            res = np.nan

        R2_serie.loc[rd, 'cor'] = res["cor"]
        R2_serie.loc[rd, 'mean_absolute_error'] = res["mean_absolute_error"]
        R2_serie.loc[rd, 'fraction_imputed'] = res["fraction_imputed"]

        print(len(masked_SNP))
        print("Result for rd {0} = cor: {1}, fraction_imputed: {2}".format(rd, res["cor"], res["fraction_imputed"] ))
    return(R2_serie)
